#include "precomp.h"
#include "ChunkHierarchy.h"
#include "ClassFactoryChunk.h"
#include "ClassFactoryMethod.h"
#include <bfc/parse/pathparse.h>
#include <bfc/parse/paramparser.h>

ChunkHierarchy::ChunkHierarchy() {
  m_maxlevel = -1;
}


ChunkHierarchy::~ChunkHierarchy() {
}


void ChunkHierarchy::addChunk(ClassFactoryChunk *chunk, const char *curpath) {
  String path = curpath;
  path.cat(StringPrintf("/%s", chunk->getClassName()));
  m_chunks.addItem(chunk, path);
  int i, n = chunk->getNumSecondaryAncestors();
  for (i=0;i<n;i++) {
    SecondaryAncestor *anc = chunk->enumSecondaryAncestor(i);
    addClass(anc->getClassName(), path);
  }
}


void ChunkHierarchy::addClass(const char *classname, const char *curpath) {
  ParamParser pp(classname, " ");
  if (pp.getNumItems() > 1)
    classname = pp.getLastItem();

  ParamParser ppp(curpath, "/");
  int i = ppp.findString(classname);
  if (i >= 0) {
#ifdef WIN32
    MessageBox(NULL, StringPrintf("Circular inheritance: %s comes up more than once in the ancestory for %s", classname, ppp.enumItem(i)), "Error", 16);
#else
    printf(StringPrintf("Circular inheritance: %s comes up more than once in the ancestory for %s", classname, ppp.enumItem(i)).getValue());
#endif
    return;
  }
  ClassFactoryChunk *chunk = chunkListInt.findItem(classname);
  if (chunk != NULL) {
    addChunk(chunk, curpath);
    return;
  }
  chunk = chunkListImp.findItem(classname);
  if (chunk != NULL) {
    addChunk(chunk, curpath);
    return;
  }
  onClassNotFound(classname, curpath);
}


int ChunkHierarchy::getNumChunks() {
  return m_chunks.getNumItems();
}


ClassFactoryChunk *ChunkHierarchy::enumChunk(int n) {
  return m_chunks.enumChunk(n);
}


const char *ChunkHierarchy::enumPath(int n) {
  return m_chunks.enumPath(n);
}


void ChunkHierarchy::reset() {
  m_chunks.removeAll();
}


void ChunkHierarchy::onClassNotFound(const char *classname, const char *hierarchypath) {
}

ClassFactoryMethod *ChunkHierarchy::findMethod(const char *methodname, String &ancestorpath, int casesensitive, ClassFactoryChunk **ptr_chunk, int mustbeimplemented, int findbasechunk) {
  int level;
  for (level=0;;level++) {
    int numprocessed = 0;
    int i, n = m_chunks.getNumItems();
    for (i=n-1;i>=0;i--) {
      const char *path = m_chunks.enumPath(i);
      PathParser pp(path);
      if (pp.getNumStrings() == level + 1) {
        numprocessed++;
        ClassFactoryChunk *chunk = m_chunks.enumChunk(i);
        int b, nb = chunk->getNumBlocks();
        for (b=0;b<nb;b++) {
          ClassFactoryBlock *block = chunk->enumBlock(b);
          if (block->getBlockType() == CF_BLOCKTYPE_METHOD) {
            ClassFactoryMethod *method = block->getMethod();
            if (mustbeimplemented && !method->isFlagSet(CF_FLAG_METHOD_IMPLEMENTED)) continue;
            if (casesensitive) {
              if (STREQL(method->getName(), methodname)) {
                if (chunk == m_chunks.enumChunk(0) && !findbasechunk) continue;
                if (ptr_chunk != NULL) *ptr_chunk = chunk;
                ancestorpath = path;
                return method;
              }
            } else {
              if (STRCASEEQL(method->getName(), methodname)) {
                if (chunk == m_chunks.enumChunk(0) && !findbasechunk) continue;
                if (ptr_chunk != NULL) *ptr_chunk = chunk;
                ancestorpath = path;
                return method;
              }
            }
          }
        }
      }
    }
    if (numprocessed == 0) break;
  }
  ancestorpath = "";
  m_maxlevel = level-1;
  return NULL;
}

ClassFactoryMethod *ChunkHierarchy::findOriginalMethod(const char *methodname, String &ancestorpath, int casesensitive, ClassFactoryChunk **ptr_chunk, int mustbeimplemented, int findbasechunk) {
  int level;
  if (m_maxlevel == -1) {
    for (level=0;;level++) {
      int numprocessed = 0;
      int i, n = m_chunks.getNumItems();
      for (i=n-1;i>=0;i--) {
        const char *path = m_chunks.enumPath(i);
        PathParser pp(path);
        if (pp.getNumStrings() == level + 1) {
          numprocessed++;
          break;
        }
      }
      if (numprocessed == 0) break;
    }
    if (level == 0) return NULL;
    level--;
    m_maxlevel = level;
  }
  level = m_maxlevel;
  for (;level>=0;level--) {
    int numprocessed = 0;
    int i, n = m_chunks.getNumItems();
    for (i=n-1;i>=0;i--) {
      const char *path = m_chunks.enumPath(i);
      PathParser pp(path);
      if (pp.getNumStrings() == level + 1) {
        numprocessed++;
        ClassFactoryChunk *chunk = m_chunks.enumChunk(i);
        int b, nb = chunk->getNumBlocks();
        for (b=0;b<nb;b++) {
          ClassFactoryBlock *block = chunk->enumBlock(b);
          if (block->getBlockType() == CF_BLOCKTYPE_METHOD) {
            ClassFactoryMethod *method = block->getMethod();
            if (mustbeimplemented && !method->isFlagSet(CF_FLAG_METHOD_IMPLEMENTED)) continue;
            if (casesensitive) {
              if (STREQL(method->getName(), methodname)) {
                if (chunk == m_chunks.enumChunk(0) && !findbasechunk) continue;
                if (ptr_chunk != NULL) *ptr_chunk = chunk;
                ancestorpath = path;
                return method;
              }
            } else {
              if (STRCASEEQL(method->getName(), methodname)) {
                if (chunk == m_chunks.enumChunk(0) && !findbasechunk) continue;
                if (ptr_chunk != NULL) *ptr_chunk = chunk;
                ancestorpath = path;
                return method;
              }
            }
          }
        }
      }
    }
    if (numprocessed == 0) break;
  }
  return NULL;
}

