#include "precomp.h"
#include "ClassFactoryChunkI.h"
#include "ClassFactoryMethodI.h"
#include "ClassFactoryDataI.h"
#include <bfc/parse/paramparser.h>

ClassFactoryChunkI::ClassFactoryChunkI(const char *_classname) : classname(_classname) {  
  m_parentChunk = NULL;
}

ClassFactoryChunkI::~ClassFactoryChunkI() { 
  blocks.deleteAll(); 
  headers.deleteAll(); 
  footers.deleteAll(); 
  errors.deleteAll(); 
  m_helpsections.deleteAll();
  secancestors.deleteAll();
}

const char *ClassFactoryChunkI::getClassName() { 
  return classname; 
}

int ClassFactoryChunkI::getNumBlocks() { 
  return blocks.getNumItems(); 
}

ClassFactoryBlock *ClassFactoryChunkI::enumBlock(int n) { 
  return blocks[n]; 
}

void ClassFactoryChunkI::addBlock(ClassFactoryBlock *b, int pos) { 
  blocks.addItem(static_cast<ClassFactoryBlockI *>(b), pos); 
  b->setParentChunk(this);
}

void ClassFactoryChunkI::removeBlock(ClassFactoryBlock *b) { 
  ClassFactoryBlockI *bb = static_cast<ClassFactoryBlockI *>(b); 
  blocks.removeItem(bb); 
  delete bb; 
}

void ClassFactoryChunkI::setClassName(const char *_classname) { 
  classname = _classname; 
}

XmlReaderParams *ClassFactoryChunkI::getParams() { 
  return &params; 
}

void ClassFactoryChunkI::setParams(XmlReaderParams *_params) { 
  for (int i=0;i<_params->getNbItems();i++) {
    const char *par = _params->getItemName(i);
    const char *val = _params->getItemValue(i);
    if (STRCASEEQL(par, "name")) continue;
    if (STRCASEEQL(par, "implementation")) continue;
    if (STRCASEEQL(par, "ancestor")) continue;
    if (STRCASEEQL(par, "inherit")) {
      secancestors.deleteAll();
      ParamParser pp(val);
      for (int i=0;i<pp.getNumItems();i++) {
        ParamParser pp2(pp.enumItem(i), " ");
        int scope = SCOPE_PRIVATE;
        String strscope = pp2.enumItem(0);
        if (pp2.getNumItems() > 1) {
          if (STRCASEEQLSAFE(strscope, "public")) scope = SCOPE_PUBLIC;
          if (STRCASEEQLSAFE(strscope, "private")) scope = SCOPE_PRIVATE;
          if (STRCASEEQLSAFE(strscope, "protected")) scope = SCOPE_PROTECTED;
        }
        addSecondaryAncestor(pp2.getLastItem(), scope);
      }
      continue;
    }
    params.addItem(par, val);
  }
}

void ClassFactoryChunkI::setRawParams(XmlReaderParams *_params) { 
  for (int i=0;i<_params->getNbItems();i++) {
    const char *par = _params->getItemName(i);
    const char *val = _params->getItemValue(i);
    if (STRCASEEQL(par, "name")) continue;
    if (STRCASEEQL(par, "implementation")) continue;
    if (STRCASEEQL(par, "ancestor")) continue;
    if (STRCASEEQL(par, "inherit")) continue;
    raw_params.addItem(par, val);
  }
}

XmlReaderParams *ClassFactoryChunkI::getRawParams() {
  return &raw_params;
}

void ClassFactoryChunkI::dropCodeBlocks(const char *btype, int methodstoo) {
  if (btype == NULL)
    footers.deleteAll();
  else
    for (int f=0;f<footers.getNumItems();f++) {
      ClassFactoryCodeBlockI *cb = footers.enumItem(f);
      const char *blocktype = cb->getParams()->getItemValue("type");
      if (btype == NULL || STRCASEEQLSAFE(btype, blocktype)) {
        footers.removeItem(cb);
        delete cb;
        f--;
      }
    }
  if (btype == NULL)
    headers.deleteAll();
  else
    for (int h=0;h<headers.getNumItems();h++) {
      ClassFactoryCodeBlockI *cb = headers.enumItem(h);
      const char *blocktype = cb->getParams()->getItemValue("type");
      if (btype == NULL || STRCASEEQLSAFE(btype, blocktype)) {
        headers.removeItem(cb);
        delete cb;
        h--;
      }
    }
  for (int i=0;i<blocks.getNumItems();i++) {
    ClassFactoryBlockI *b = blocks.enumItem(i);
    switch (b->getBlockType()) {
      case CF_BLOCKTYPE_CODEBLOCK: {
        const char *blocktype = b->getCodeBlock()->getParams()->getItemValue("type");
        if (btype != NULL && !STRCASEEQLSAFE(btype, blocktype)) continue;
        blocks.removeByPos(i);
        delete b;
        i--;
        break;
      }
      case CF_BLOCKTYPE_METHOD:
        if (methodstoo) {
          b->getMethod()->dropCodeBlocks(btype);
        }
        break;
      }
  }
  if (btype == NULL || STRCASEEQLSAFE("h", btype)) 
    params.removeItem("kw");
}

int ClassFactoryChunkI::getBlockNum(ClassFactoryBlock *block) {
  ClassFactoryBlockI *pb = static_cast<ClassFactoryBlockI *>(block);
  for (int i=0;i<blocks.getNumItems();i++)
    if (blocks.enumItem(i) == pb) return i;
  return -1;
}

void ClassFactoryChunkI::setImplementationName(const char *_implname) {
  implname = _implname;
}

const char *ClassFactoryChunkI::getImplementationName() {
  if (classname.isempty() && implname.isempty()) return NULL;
  if (implname.isempty()) {
    if (autoimplname.isempty())
      autoimplname = StringPrintf("%sI", getClassName());
    return autoimplname;
  }
  return implname;
}

ClassFactoryCodeBlock *ClassFactoryChunkI::getHeader(const char *type) {
  foreach(headers)
    ClassFactoryCodeBlockI *cb = headers.getfor();
    const char *cbt = cb->getParams()->getItemValue("type");
    if (STRCASEEQLSAFE(cbt, type)) return cb;
  endfor;
  return NULL;
}

ClassFactoryCodeBlock *ClassFactoryChunkI::getFooter(const char *type) {
  foreach(footers)
    ClassFactoryCodeBlockI *cb = footers.getfor();
    const char *cbt = cb->getParams()->getItemValue("type");
    if (STRCASEEQLSAFE(cbt, type)) return cb;
  endfor;
  return NULL;
}

void ClassFactoryChunkI::setHeader(ClassFactoryCodeBlock *block) {
  const char *type = block->getParams()->getItemValue("type");
  foreach(headers)
    ClassFactoryCodeBlockI *cb = headers.getfor();
    const char *cbt = cb->getParams()->getItemValue("type");
    if (STRCASEEQLSAFE(cbt, type)) { headers.removeItem(cb); delete cb; break; }
  endfor;
  headers.addItem(static_cast<ClassFactoryCodeBlockI*>(block));
}

void ClassFactoryChunkI::setFooter(ClassFactoryCodeBlock *block) {
  const char *type = block->getParams()->getItemValue("type");
  foreach(footers)
    ClassFactoryCodeBlockI *cb = footers.getfor();
    const char *cbt = cb->getParams()->getItemValue("type");
    if (STRCASEEQLSAFE(cbt, type)) { footers.removeItem(cb); delete cb; break; }
  endfor;
  footers.addItem(static_cast<ClassFactoryCodeBlockI*>(block));
}

ClassFactoryMethod *ClassFactoryChunkI::getConstructor(int n) {
  int _n = 0;
  foreach(blocks)
    if (blocks.getfor()->getBlockType() == CF_BLOCKTYPE_METHOD) {
      ClassFactoryMethod *m = blocks.getfor()->getMethod();
      if (m != NULL) {
        const char *mname = m->getName();
        if (STRCASEEQL(mname, getClassName()) && _n++ == n) return m;
      }
    }
  endfor;
  return NULL;
}

ClassFactoryMethod *ClassFactoryChunkI::getDestructor(int n) {
  int _n = 0;
  foreach(blocks)
    if (blocks.getfor()->getBlockType() == CF_BLOCKTYPE_METHOD) {
      ClassFactoryMethod *m = blocks.getfor()->getMethod();
      if (m != NULL) {
        const char *mname = m->getName();
        if (STRCASEEQL(mname, StringPrintf("~%s", getClassName())) && _n++ == n) return m;
      }
    }
  endfor;
  return NULL;
}

void ClassFactoryChunkI::setFlag(int flag) { 
  flag_setFlag(flag); 
}

int ClassFactoryChunkI::isFlagSet(int flag) { 
  return flag_isFlagSet(flag); 
}

void ClassFactoryChunkI::clearFlag(int flag) { 
  flag_clearFlag(flag); 
}

void ClassFactoryChunkI::addErrorMessage(const char *msg, int col) { 
  errors.addItem(new String(msg)); 
  errcols.addItem(col); 
}

int ClassFactoryChunkI::getNumErrors() { 
  return errors.getNumItems(); 
}

const char *ClassFactoryChunkI::enumError(int n, int *col) { 
  *col = errcols.enumItem(n); 
  return errors.enumItem(n)->getValue(); 
}

void ClassFactoryChunkI::clearErrors() { 
  errors.deleteAll(); 
}

void ClassFactoryChunkI::setAncestor(const char *_ancestor) { 
  ancestor = _ancestor; 
}

const char *ClassFactoryChunkI::getAncestor() { 
  return ancestor; 
}

void ClassFactoryChunkI::copy(ClassFactoryChunk *c) {
  ClassFactoryChunkI *_c = static_cast<ClassFactoryChunkI *>(c);
  while (params.getNbItems()) {
    params.removeItem(params.getItemName(0));
  }
  XmlReaderParams *par = c->getParams();
  int p, np = par->getNbItems();
  for (p=0;p<np;p++) {
    params.addItem(par->getItemName(p), par->getItemValue(p));
  }
  par = c->getRawParams();
  np = par->getNbItems();
  for (p=0;p<np;p++) {
    raw_params.addItem(par->getItemName(p), par->getItemValue(p));
  }
  blocks.deleteAll(); 
  headers.deleteAll(); 
  footers.deleteAll(); 
  errors.deleteAll();
  errcols.removeAll();
  secancestors.deleteAll();
  flag_copy(static_cast<ClassFactoryChunkI*>(c));
  autoimplname = "";
  classname = c->getClassName();
  setClassKW(c->getClassKW());
  implname = (static_cast<ClassFactoryChunkI*>(c))->implname;
  int a, na = c->getNumSecondaryAncestors();
  for (a = 0;a<na;a++) {
    SecondaryAncestor *anc = c->enumSecondaryAncestor(a);
    addSecondaryAncestor(anc->getClassName(), anc->getScope());
  }
  int b, nb = c->getNumBlocks();
  for (b=0;b<nb;b++) {
    ClassFactoryBlockI *block = new ClassFactoryBlockI();
    block->copy(c->enumBlock(b));
    addBlock(block);
  }
  int h, nh = _c->headers.getNumItems();
  for (h=0;h<nh;h++) {
    ClassFactoryCodeBlockI *block = new ClassFactoryCodeBlockI();
    block->copy(_c->headers.enumItem(h));
    headers.addItem(block);
  }
  int f, nf = _c->footers.getNumItems();
  for (f=0;f<nf;f++) {
    ClassFactoryCodeBlockI *block = new ClassFactoryCodeBlockI();
    block->copy(_c->footers.enumItem(f));
    footers.addItem(block);
  }
  m_helpsections.deleteAll();
  int hs, nhs = _c->getNumHelpSections();
  for (hs=0;hs<nhs;hs++) {
    HelpSectionI *section = new HelpSectionI();
    section->clone(_c->enumHelpSection(hs));
    m_helpsections.addItem(section);
  }
  setAncestor(c->getAncestor());
}

void ClassFactoryChunkI::clearAllFlags(int methodstoo) {
  flag_clearAllFlags();
  if (methodstoo) {
    foreach(blocks)
      ClassFactoryBlock *block = blocks.getfor();
      if (block->getBlockType() == CF_BLOCKTYPE_METHOD) {
       block->getMethod()->clearFlag(-1);
      }
      if (block->getBlockType() == CF_BLOCKTYPE_DATA) {
       block->getData()->clearFlag(-1);
      }
    endfor;
  }
}

int ClassFactoryChunkI::haveCodeBlocks(const char *btype) {
  int n = 0;
  if (getFooter(btype)) n++;
  if (getHeader(btype)) n++;
  foreach(blocks)
    if (blocks.getfor()->getBlockType() == CF_BLOCKTYPE_CODEBLOCK) 
      if (STRCASEEQL(blocks.getfor()->getCodeBlock()->getType(), btype))
        n++;
  endfor
  return n;
}

void ClassFactoryChunkI::deleteBlock(ClassFactoryBlock *block) {
  ClassFactoryBlockI *pblock = static_cast<ClassFactoryBlockI *>(block);
  foreach(blocks)
    ClassFactoryBlockI *thisblock = blocks.getfor();
    if (thisblock == pblock) {
      delete pblock;
      blocks.removeItem(pblock);
      return;
    }
  endfor;
}

void ClassFactoryChunkI::setClassKW(const char *_kw) { 
  params.removeItem("kw"); 
  params.addItem("kw", _kw); 
}

const char *ClassFactoryChunkI::getClassKW() { 
  return params.getItemValue2("kw"); 
}

int ClassFactoryChunkI::isConstructor(ClassFactoryMethod *m) {
  if (STRCASEEQLSAFE(m->getName(), "ctor")) return 1;
  if (STRCASEEQLSAFE(getClassName(), m->getName())) return 1;
  if (STRCASEEQLSAFE(getImplementationName(), m->getName())) return 1;
  return 0;
}

int ClassFactoryChunkI::isDestructor(ClassFactoryMethod *m) {
  if (STRCASEEQLSAFE(m->getName(), "dtor")) return 1;
  if (STRCASEEQLSAFE(StringPrintf("~%s", getClassName()), m->getName())) return 1;
  if (STRCASEEQLSAFE(StringPrintf("~%s", getImplementationName()), m->getName())) return 1;
  return 0;
}

void ClassFactoryChunkI::setClassType(int ct) { 
  params.removeItem("class_type");
  params.addItem("class_type", StringPrintf("%d", ct));
}

int ClassFactoryChunkI::getClassType() { 
  return params.getItemValueInt("class_type"); 
}

int ClassFactoryChunkI::getNumSecondaryAncestors() {
  return secancestors.getNumItems();
}

SecondaryAncestor *ClassFactoryChunkI::enumSecondaryAncestor(int n) {
  return secancestors.enumItem(n);
}

void ClassFactoryChunkI::removeSecondaryAncestor(const char *anc) {
  foreach(secancestors)
    SecondaryAncestorI *s = secancestors.getfor();
    if (STRCASEEQL(s->getClassName(), anc)) {
      delete s;
      secancestors.removeByPos(foreach_index);
      return;
    }
  endfor;
}

void ClassFactoryChunkI::removeSecondaryAncestorNum(int n) {
  SecondaryAncestorI *s = secancestors.enumItem(n);
  if (s != NULL) {
    delete s;
    secancestors.removeByPos(n);      
  }
}

void ClassFactoryChunkI::addSecondaryAncestor(const char *anc, int scope) {
  removeSecondaryAncestor(anc);
  secancestors.addItem(new SecondaryAncestorI(anc, scope));
}

int ClassFactoryChunkI::moveBlock(ClassFactoryBlock *block, int newpos) { // returns old pos
  int curpos = getBlockNum(block);
  if (curpos == newpos) return curpos;
  ClassFactoryBlockI *b = static_cast<ClassFactoryBlockI *>(block);
  blocks.addItem(b, newpos);
  if (curpos > newpos) curpos++;
  blocks.removeByPos(curpos);
  return curpos;
}

void ClassFactoryChunkI::addHelpSection(HelpSection *section) {
  m_helpsections.addItem(static_cast<HelpSectionI*>(section));
}

int ClassFactoryChunkI::getNumHelpSections() {
  return m_helpsections.getNumItems();
}

HelpSection *ClassFactoryChunkI::enumHelpSection(int n) {
  return m_helpsections.enumItem(n);
}

void ClassFactoryChunkI::deleteHelpSection(HelpSection *s) {
  m_helpsections.removeItem(static_cast<HelpSectionI*>(s));
  delete static_cast<HelpSectionI*>(s);
}

void ClassFactoryChunkI::setParentChunk(ClassFactoryChunk *chunk) {
  m_parentChunk = chunk;
}

ClassFactoryChunk *ClassFactoryChunkI::getParentChunk() {
  return m_parentChunk;
}
