#include "precomp.h"
#include <bfc/string/stringdict.h>
#include <bfc/parse/paramparser.h>
#include "../classhierarchy/Project.h"
#include "../classhierarchy/ClassFactoryChunk.h"
#include "../classhierarchy/ClassFactoryMethod.h"
#include "../classhierarchy/ClassFactoryData.h"
#include "../classhierarchy/ClassFactoryCodeBlock.h"
#include "../classhierarchy/ClassFactoryBlock.h"
#include "../classhierarchy/ClassFactorySubChunk.h"
#include "../classhierarchy/XHDProcessor.h"
#include "../classhierarchy/SecondaryAncestor.h"
#include "../outputfactories/CPPToXHDMerge.h"
#include "ImplementationFactory.h"
#include <api/xml/xmlparams.h>
#include "../classhierarchy/Workspace.h"

static const char *FORMAT_NAME = "Implementation";
static const char *FORMAT_DESC = "Implementation class factory";

ImplementationFactory::ImplementationFactory() : chunk(NULL), m_ppercent(NULL) {
}

ImplementationFactory::~ImplementationFactory() {
}

int ImplementationFactory::tryFormat(const char *format) {
  return (STRCASEEQLSAFE(format, FORMAT_NAME));
}

const char *ImplementationFactory::getName() {
  return FORMAT_NAME;
}

const char *ImplementationFactory::getDescription() {
  return FORMAT_DESC;
}

int ImplementationFactory::getNumModules() {
  return 2;
}

const char *ImplementationFactory::enumModuleName(int i) {
  switch (i) {
    case 0: return headerName;
    case 1: return moduleName;
  }
  return NULL;
}

const char *ImplementationFactory::enumModuleData(int i) {
  switch (i) {
    case 0: return headerData;
    case 1: return moduleData;
  }
  return NULL;
}

void ImplementationFactory::setChunk(ClassFactoryChunk *_chunk) {
  OutputFactoryBase::setChunk(_chunk);
  chunk = _chunk;
  m_ppercent = NULL;

  implclassname = chunk->getImplementationName();
  interfclassname = chunk->getClassName();
  ancestorName = chunk->getAncestor();

  XmlReaderParams *par = chunk->getParams();

  headerName = par->getItemValue("class_header");
  if (headerName.isempty()) {
    headerName = implclassname + ".h";
    par->removeItem("class_header");
    par->addItem("class_header", headerName);
  }

  moduleName = par->getItemValue("class_module");
  if (moduleName.isempty()) {
    moduleName = implclassname + ".cpp";
    par->removeItem("class_module");
    par->addItem("class_module", moduleName);
  }

  headerData = "";
  moduleData = "";
}

int ImplementationFactory::process() {
  if (chunk == NULL) return 0;
  return processImplementation(chunk, NULL);
}

void ImplementationFactory::setProgressCallback(ProgressCallback *cb) { m_ppercent = cb; }


int ImplementationFactory::processImplementation(ClassFactoryChunk *chunk, PtrList < String > *additional_members) {
  char ucclassname[256];
  char lcclassname[256];

  ClassFactoryChunk *parentchunk = chunk->getParentChunk();
  int issubchunk = parentchunk != NULL;

  implclassname = chunk->getImplementationName();
  interfclassname = chunk->getClassName();

  // blow out the old data.
  moduleData = "";
  headerData = "";

  str.reset();

  String thisclassname;
  thisclassname = chunk->getImplementationName();

  STRCPY(ucclassname, thisclassname);
  STRCPY(lcclassname, ucclassname);
  STRTOUPPER(ucclassname);
  STRTOLOWER(lcclassname);

  String fact = chunk->getParams()->getItemValue("xhd_factories");
  ParamParser ppf(fact, ";", 1);
  int interfaced = ppf.hasString("Interface");
  int dispatched = chunk->getParams()->getItemValueInt("dispatchable");

  String friendstring;
  String friends = chunk->getParams()->getItemValue("friends");
  XHDProcessor preproc(0, chunk);
  int cln = preproc.getNumFactories();
  int icl = 0;
  for (icl = 0; icl<cln; icl++) {
    ProcessorEntry *pe = preproc.enumProcessorEntry(icl);
    if (pe->m_needfriend && !ppf.hasString(pe->m_classname)) {
      friendstring += "friend ";
      friendstring += pe->m_classname;
      friendstring += ";\n";
    }
  }

  if (!issubchunk && !chunk->isFlagSet(CF_FLAG_HAS_H_CODEBLOCKS)) {
    str += "#ifndef __";
    str += ucclassname;
    str += "_IMPL_H\n";
    str += "#define __";
    str += ucclassname;
    str += "_IMPL_H\n";
    str += "\n";

    int scripted = 0;
    for (int pf=0;pf<ppf.getNumItems();pf++) {
      if (!STRNICMP(ppf.enumItem(pf), "Script", 6))
        scripted = 1;
    }

    // todo: make factories record their .h's in the structure
    str += "/*<?<autoheader/>*/\n/*?>*/\n\n";
  }

  if (chunk->getHeader("h")) {
    String x = chunk->getHeader("h")->getCodeBlock();
    if (!x.isempty()) {
      processInlineXml(chunk, NULL, x, getWorkspace());
      str += x;
    }
  }

  str += "class ";
  String kw = chunk->getClassKW();
  if (!kw.isempty()) {
    str += kw;
    str += " ";
  }

  str += thisclassname;
  int gotcol = 0;

  if (!chunk->getAncestor()) {
    if (interfaced) {
      str += " : public ";
      str += chunk->getClassName();
      gotcol = 1;
    }
  } else {
    str += " : public ";
    str += chunk->getAncestor();
    gotcol = 1;
  }

  for (int hh=0;hh<chunk->getNumSecondaryAncestors();hh++) {
    SecondaryAncestor *anc = chunk->enumSecondaryAncestor(hh);
    const char *s = anc->getClassName();
    ClassFactoryChunk *chunk = getWorkspace()->findChunk(s);
    if (!gotcol) {
      str += " : ";
      gotcol = 1;
    } else
      str += ", ";
    switch (anc->getScope()) {
      case SCOPE_PUBLIC:
        str += "public ";
        break;
      case SCOPE_PRIVATE:
        str += "private ";
        break;
      case SCOPE_PROTECTED:
        str += "protected ";
        break;
    }
    str += chunk ? chunk->getImplementationName() : s;
  }

  // use getsecondaryancestors on chunk instead to fix
  String custanc = chunk->getParams()->getItemValue("ancestors");
  ParamParser pp2(custanc);
  for (int h=0;h<pp2.getNumItems();h++) {
    const char *s = pp2.enumItem(h);
    if (!gotcol) {
      str += " : ";
      gotcol = 1;
    } else
      str += ", ";
    str += s;
  }

  int lastwrittenscope = SCOPE_PRIVATE;

  str += " {";//\n";
  if (!chunk->isFlagSet(CF_FLAG_HAS_H_CODEBLOCKS))
    str += "\n";

  ClassFactoryCodeBlock *hdr = chunk->getHeader("class");
  if (hdr) {

    String x = hdr->getCodeBlock();
    if (!x.isempty()) {
      processInlineXml(chunk, NULL, x, getWorkspace());
      str += x;
      CPPToXHDMerge::updateScope(x, &lastwrittenscope);
    }
  }

  str += friendstring;

  int ds = 0;

  int x = 0;
  for (x=0;x<2;x++) {
    int lastwascode = 1;
    if (chunk->getNumBlocks() > 0) {
      //if (ds) str += "\n";
      int mc = 0;
      int hadconst = 0;
      int haddest = 0;
      int nb = chunk->getNumBlocks();
      for (int i=0;i<nb;i++) {
        if (m_ppercent) m_ppercent->setPercent((int)(((x*0.25) + ((float)i / (float)nb) / 4) * 100));
        ClassFactoryBlock *b = chunk->enumBlock(i);
        switch (b->getBlockType()) {
          case CF_BLOCKTYPE_METHOD: {
            int isinline = 0;
            ClassFactoryMethod *m = b->getMethod();

            int scriptable = m->getParams()->getItemValueInt("scriptable");
            int isevent = m->getParams()->getItemValueInt("isevent");

            int where = 0;
            int isi = m->isInline(&where);
            if (x == 1 && !isi) continue;
            if (x == 1 && isi) {
              if (where != x) continue;
              isinline = 1;
            }
            int isconstructor = chunk->isConstructor(m);
            int isdestructor = chunk->isDestructor(m);

            if (isconstructor && (mc == 0 || mc == 1)) { hadconst = 1; };
            if (isdestructor && (mc == 0 || mc == 1)) { haddest = 1; };
            //if (mc == 2 && hadconst && haddest && !lastwascode) str += "\n";
            mc++;
            ClassFactoryCodeBlock *block = m->getCodeBlock(isinline ? "inline" : "h");
            ClassFactoryCodeBlock *cbp = m->getCodeBlock("cpp");
            int needmember = (isconstructor || isdestructor || block != NULL || cbp != NULL || !m->isFlagSet(CF_FLAG_METHOD_IMPLEMENTED));
            if (block && cbp) {
              //chunk->addErrorMessage(StringPrintf("%s has a body in both .cpp and .h file", m->getName()));
              // warning !
            }
            if (!needmember && additional_members != NULL) { // check in our custom list
              int im, imn = additional_members->getNumItems();
              for (im=0;im<imn;im++) {
                if (STRCASEEQL(additional_members->enumItem(im)->getValue(), m->getName())) {
                  needmember = 1;
                  break;
                }
              }
            }
            int curscope = m->getScope();
            if (needmember) {
              if (x == 0 && lastwrittenscope != curscope) {
                str += ""; // indent
                switch (curscope) {
                  case SCOPE_PRIVATE:
                    str += "private:";
                    break;
                  case SCOPE_PUBLIC:
                    str += "public:";
                    break;
                  case SCOPE_PROTECTED:
                    str += "protected:";
                    break;
                  }
                str += "\n";
                lastwrittenscope = curscope;
              }
              int start = str.getLineCount();
              m->setStartLine(start, x == 1 ? "inline" : "h");
              String _s;
              String prefix = (x == 1) ? (thisclassname + "::").getValue() : "";
              methodToText(chunk, m, _s, 1, prefix, thisclassname, 1);
              if (x == 0)
                str += "  "; // indent

              if (x == 0 && dispatched && interfaced) {
                if (m->getParams()->getItemValueInt("nodispatch")) {
                  str += "NODISPATCH ";
                } else if (m->getParams()->getItemValueInt("dispatch") || m->getParams()->getItemValueInt("dispatch_id", -1) != -1) {
                  if (m->getParams()->getItemValueInt("dispatch_id", -1) != -1) {
                    str += StringPrintf("DISPATCH(%s) ", m->getParams()->getItemValue("dispatch_id"));
                  } else {
                    str += "DISPATCH_ ";
                  }
                }
              }

              if (x == 0 && scriptable) {
                str += "SCRIPT ";
              }

              if (x == 0 && isevent) {
                str += "EVENT ";
              }

              if (x == 0 && m->isVirtual()) str += "virtual ";
              if (x == 0 && m->isStatic()) str += "static ";
              if (isi && x == where) str += "inline ";
              str += _s;
              String mod = m->getPostDeclModifier();
              if (x == 0) str += mod;
              int purevirt=0;
              if (m->isVirtual(&purevirt) && purevirt) {
                str += "=0;";
              }
              const char *hh = m->getHeader(isinline ? "inline" : "h");
              str += hh;
              if (block != NULL && ((x == 0 && !isi) || (isi && where == x))) {
                if (!hh && mod.isempty())
                  str += " ";
                str += "{";
  //              if (!ATOI(m->getParams()->getItemValue("onlinebrace")))
  //                str += "\n";
                String x = block->getCodeBlock();
                if (!x.isempty()) {
                  processInlineXml(chunk, m, x, getWorkspace());
                  str += x;
                }
                str += "}"; //\n";
                String f = m->getFooter(isinline ? "inline" : "h");
                if (!f.isempty()) {
                  processInlineXml(chunk, m, f, getWorkspace());
                  str += f;
                } else
                  str += "\n";
              } else {
                if (!purevirt) str += ";"; //\n";
                String f = m->getFooter(isinline ? "inline" : "h");
                if (!f.isempty()) {
                  processInlineXml(chunk, m, f, getWorkspace());
                  str += f;
                } else
                  str += "\n";
              }

              m->setNumLines(str.getLineCount() - start, x == 1 ? "inline" : "h");
              lastwascode = 0;
            }
            break;
          }
          case CF_BLOCKTYPE_CODEBLOCK: {
            ClassFactoryCodeBlock *cb = b->getCodeBlock();
            const char *t = cb->getParams()->getItemValue("type");
            if (t == NULL) t = "cpp";
            if (STRCASEEQL(t, x == 1 ? "inline" : "h")) {
              String x = cb->getCodeBlock();
              if (!x.isempty()) {
                processInlineXml(chunk, NULL, x, getWorkspace());
                str += x;
              }
              CPPToXHDMerge::updateScope(cb->getCodeBlock(), &lastwrittenscope);
              lastwascode = 1;
            }
            break;
          }
          case CF_BLOCKTYPE_DATA: {
            if (x != 0) continue;
            ClassFactoryData *data = b->getData();
            String s;
            s += "  ";
            if (data->isStatic())
              s += "static ";
            s += data->getDataType();
            if (!(s.lastChar() == '*' || s.lastChar() == '&'))
              s += " ";
            s += data->getName();
            s += ";\n";
            str += s;
            break;
          }
          case CF_BLOCKTYPE_SUBCHUNK: {
            if (x != 0) continue;
            ClassFactorySubChunk *subchunk = b->getSubChunk();
            ImplementationFactory sif;
            sif.setChunk(subchunk->getChunk());
            sif.process();
            str += sif.enumModuleData(0);
          }
        }
      }
    }

    if (x == 0) {
      str += "};";
      if (chunk->getFooter("class")) {
        String x = chunk->getFooter("class")->getCodeBlock();
        if (!x.isempty()) {
          processInlineXml(chunk, NULL, x, getWorkspace());
          str += x;
        }
      }
    }

    if (!chunk->isFlagSet(CF_FLAG_HAS_H_CODEBLOCKS))
      str += "\n";
  }

  if (chunk->getFooter("h")) {
    String x = chunk->getFooter("h")->getCodeBlock();
    if (!x.isempty()) {
      processInlineXml(chunk, NULL, x, getWorkspace());
      str += x;
    }
  }

  if (!issubchunk && !chunk->isFlagSet(CF_FLAG_HAS_H_CODEBLOCKS)) {
    str += "\n";
    str += "#endif // __";
    str += ucclassname;
    str += "_IMPL_H\n";
  }

  headerData = str;

  String hdrName;
  str = "";

  hdrName = chunk->getImplementationName(); hdrName += ".h";

  const char *afx = chunk->getParams()->getItemValue("stdafx");
  if (!issubchunk && !chunk->isFlagSet(CF_FLAG_HAS_CPP_CODEBLOCKS)) {
    if (afx && ATOI(afx) != 0)
      str += "#include \"precomp.h\"\n";

    String x = "//<?#include \"<class data=\"implementationheader\"/>\"\n";
    x += "//?>\n\n";
    processInlineXml(chunk, NULL, x, getWorkspace());
    str += x;
  }

  if (chunk->getHeader("cpp")) {
    String x = chunk->getHeader("cpp")->getCodeBlock();
    if (!x.isempty()) {
      processInlineXml(chunk, NULL, x, getWorkspace());
      str += x;
    }
  }

  int lastwascode = 1;

  int nb = chunk->getNumBlocks();
  for (int i=0;i<nb;i++) {
    if (m_ppercent) m_ppercent->setPercent((int)((0.50 + ((float)i / (float)nb) / 2) * 100));
    ClassFactoryBlock *b = chunk->enumBlock(i);
    switch (b->getBlockType()) {
      case CF_BLOCKTYPE_METHOD: {
        ClassFactoryMethod *m = b->getMethod();
        ClassFactoryCodeBlock *cb = m->getCodeBlock("cpp");
        ClassFactoryCodeBlock *cbh = m->getCodeBlock("h");

        if (cb == NULL) continue;

        int isconstructor = chunk->isConstructor(m);
        int isdestructor = chunk->isDestructor(m);

//          if ((isconstructor || isdestructor) && z == 0) continue;

        int needmember = (cbh == NULL && ((isconstructor || isdestructor || cb != NULL || !m->isFlagSet(CF_FLAG_METHOD_IMPLEMENTED))));
        if (!needmember && additional_members != NULL) { // check in our custom list
          int im, imn = additional_members->getNumItems();
          for (im=0;im<imn;im++) {
            if (STRCASEEQL(additional_members->enumItem(im)->getValue(), m->getName())) {
              needmember = 1;
              break;
            }
          }
        }
        if (needmember) {
          if (!lastwascode) str += "\n";
          int start = str.getLineCount();
          m->setStartLine(start, "cpp");
          String _str;
          writeImplementationMethod(0, chunk, m, _str, this, NULL, getWorkspace());
          str += _str;
          m->setNumLines(str.getLineCount() - start, "cpp");
          lastwascode = 0;
       }
        break;
      }
      case CF_BLOCKTYPE_CODEBLOCK: {
        ClassFactoryCodeBlock *cb = b->getCodeBlock();
        const char *t = cb->getParams()->getItemValue("type");
        if (t == NULL) t = "cpp";
        if (STRCASEEQL(t, "cpp")) {
          String x = cb->getCodeBlock();
          if (!x.isempty()) {
            processInlineXml(chunk, NULL, x, getWorkspace());
            str += x;
            lastwascode = 1;
          }
        }
        break;
      }
      case CF_BLOCKTYPE_DATA: {
        ClassFactoryData *data = b->getData();
        if (!data->isStatic()) break;
        String s;
        s += data->getDataType();
        if (!(s.lastChar() == '*' || s.lastChar() == '&'))
          s += " ";
        s += chunk->getImplementationName();
        s += "::";
        s += data->getName();
        String init = data->getInitValue();
        if (!init.isempty())
          s += String(" = ") + init;
        s += ";\n";
        str += s;
      }
      case CF_BLOCKTYPE_SUBCHUNK: {
        ClassFactorySubChunk *subchunk = b->getSubChunk();
        ImplementationFactory sif;
        sif.setChunk(subchunk->getChunk());
        sif.process();
        str += sif.enumModuleData(1);
      }
    }
  }

  if (chunk->getFooter("cpp")) {
    String x = chunk->getFooter("cpp")->getCodeBlock();
    if (!x.isempty()) {
      processInlineXml(chunk, NULL, x, getWorkspace());
      str += x;
    }
  }

  moduleData = str;

  return 1;
}

int ImplementationFactory::process_add_members(const char *format, PtrList < String > *members) {
  if (chunk == NULL) return 0;
  if (STRCASEEQLSAFE(format, FORMAT_NAME)) return processImplementation(chunk, members);
  return 0;
}

void ImplementationFactory::writeImplementationMethod(int helper, ClassFactoryChunk *chunk, ClassFactoryMethod *m, String &str, OutputFactoryBase *base, const char *forcecallclass, Workspace *workspace, int bodyonly) {
  if (0) {
    if (!ATOI(m->getParams()->getItemValue("implementation")))
      str += StringPrintf("///> Implementation for %s\n", chunk->getClassName());
  }
  ClassFactoryCodeBlock *cb = m->getCodeBlock("cpp");
  ClassFactoryCodeBlock *cbh = m->getCodeBlock("h");
  if (0) { // fucko
    if (cb == NULL && cbh != NULL) {
      cb = cbh;
      cbh = NULL;
    }
  }

  String thisclassname = chunk->getImplementationName();

  String clsprefx;
  clsprefx += thisclassname;
  clsprefx += "::";
  String _s;
  if (!bodyonly) {
    _methodToText(chunk, m, _s, 0, clsprefx, thisclassname, 1);
    str += _s;
    String hcpp = m->getHeader("cpp");
    if (!hcpp.isempty()) {
      processInlineXml(chunk, m, hcpp, workspace);
      str += hcpp;
    } else
      str += " ";
    str += "{";
  }
  if (!helper && cb != NULL && cb->getCodeBlock() != NULL) {
    String x = cb->getCodeBlock();
    if (!x.isempty()) {
      processInlineXml(chunk, m, x, workspace);
      str += x;
    }
  } else { // if there is no implementation yet, we've created one, so call the parent class if needed
    str += "\n";
    if (!chunk->isConstructor(m) && !chunk->isDestructor(m) && m->isVirtual() /*forcecallclass != NULL || m->isFlagSet(CF_FLAG_METHOD_IMPLEMENTED)*/) { // if this method has already been implemented by one of our ancestors, we need to call it
      String stx;
      String rt = m->getReturnType();
      int calluptype = m->getParams()->getItemValueInt("calluptype");
      if (!rt.isempty() && !rt.iscaseequal("void")) {
        if (calluptype == 0) {
          stx += "  //<? <callup><rettype/> _retVal = <ancestor/>::<name/>(<params/>);</callup><nocallup><rettype/> _retVal = <defretval/>;</nocallup>";
          stx += base ? base->CRLF : "\n";
          stx += "  //?>";
          stx += base ? base->CRLF : "\n";
        } else {
          // nothing to do
        }
      } else {
        if (calluptype == 0) {
          stx += "  //<? <callup><ancestor/>::<name/>(<params/>);</callup>";
          stx += base ? base->CRLF : "\n";
          stx += "  //?>";
          stx += base ? base->CRLF : "\n";
        }
      }
      if (!stx.isempty())
        processInlineXml(chunk, m, stx, workspace);
      str += stx;
      str += base ? base->CRLF : "\n";
      str += "  // TODO: add custom code here";
      str += base ? base->CRLF : "\n";
      str += base ? base->CRLF : "\n";
      if (!rt.isempty() && !rt.iscaseequal("void")) {
        if (calluptype == 0) {
          stx.trunc(0);
          stx += "  //<? return _retVal;";
          stx += base ? base->CRLF : "\n";
          stx += "  //?>";
          stx += base ? base->CRLF : "\n";
          processInlineXml(chunk, m, stx, workspace);
          str += stx;
        } else {
          stx += "  //<? <callup>return <ancestor/>::<name/>(<params/>);</callup><nocallup>return <defretval/>;</nocallup>";
          stx += base ? base->CRLF : "\n";
          stx += "  //?>";
          stx += base ? base->CRLF : "\n";
        }
      } else {
        if (calluptype == 0) {
        } else {
          stx += "  //<? <callup><ancestor/>::<name/>(<params/>);</callup>";
          stx += base ? base->CRLF : "\n";
          stx += "  //?>";
          stx += base ? base->CRLF : "\n";
        }
      }
    }
  }
  if (!bodyonly) {
    str += "}"; //\n";

    String f = m->getFooter("cpp");
    if (!f.isempty()) {
      processInlineXml(chunk, m, f, workspace);
      str += f;
    } else
      str += "\n";
  }
}





