#include "precomp.h"
#include <sys/stat.h>
#include "InlineInterpreter.h"
#include <bfc/string/string.h>
#include "../classhierarchy/ClassFactoryChunkI.h"
#include "../classhierarchy/ClassFactoryBlock.h"
#include "../classhierarchy/ClassFactoryMethod.h"
#include "../classhierarchy/ClassFactoryParam.h"
#include "../classhierarchy/ClassFactoryParser.h"
#include "../outputfactories/CPPToXHDMerge.h"
#include <bfc/parse/pathparse.h>
#include <bfc/parse/paramparser.h>
#include "../classhierarchy/Project.h"
#include <bfc/util/profiler.h>

#define MAXBUF 16 * 1024

extern "C" unsigned long int crc32file(const char *name);


InlineInterpreter::InlineInterpreter(const char *pathfile, ClassFactoryChunk *chunk, int rewrite) {
  process(pathfile, chunk, rewrite);
}

InlineInterpreter::InlineInterpreter() {
}

InlineInterpreter::~InlineInterpreter() {
}

void InlineInterpreter::scanXml(const char *p, int *inxml) {
  while (*p) {
    if (!STRNCMP(p, "//<?", 4)) (*inxml)++;
    if (!STRNCMP(p, "//?>", 4)) (*inxml)--;
    if (!STRNCMP(p, "/*<?", 4)) (*inxml)++;
    if (!STRNCMP(p, "?>*/", 4)) (*inxml)--;
    p++;
  }
}

void InlineInterpreter::getBlock(char *dest, int n, VFILE *fl, int *bracketcount) {
  char *p = dest;
  int size = n;
  int inxml = 0;
  do {
    CPPToXHDMerge::Vfgets_cpp(p, size, fl, NULL, NULL, bracketcount);
    scanXml(p, &inxml);
    size -= STRLEN(p);
    p += STRLEN(p);
  } while (inxml > 0 && size > 0);
}

int InlineInterpreter::process(const char *pathfile, ClassFactoryChunk *chunk, int rewrite) {
  PR_ENTER("InlineInterpreter::process");
  VFILE *inf = Vfopen(pathfile, "rt", TRUE);

  if (!inf) return 0;

  String outfile;
  String path;
  PathParser pp(pathfile);
  if (Std::isDirChar(*pathfile)) path += DIRCHARSTR;
  for (int i=0;i<pp.getNumStrings()-1;i++) { path += pp.enumString(i); path += DIRCHARSTR; }
  String filename = pp.getLastString();
  ParamParser parp(filename, ".");
  filename = parp.enumItem(0);
  outfile = path + filename + "._tmp_";

  VFILE *outf = NULL;

  if (rewrite) {
    outf = Vfopen(outfile, "w+t", TRUE);
    if (!outf) { Vfclose(inf); return 0; }
  }

  Vfseek(inf, 0, SEEK_END);
  int size = Vftell(inf);
  Vfseek(inf, 0, SEEK_SET);

  char line[MAXBUF];
  int linecount = 0;
  int gbracecount = 0;
  int gotmethod = 0;
  int isinclass = 0;

  ClassFactoryMethod *currentmethod = NULL;
  ClassFactoryMethod *lastmethodparsed = NULL;

  while (!Vfeof(inf)) {
    getBlock(line, MAXBUF-1, inf, &gbracecount);
    linecount++;

    if (!currentmethod) {

      if (isinclass || CPPToXHDMerge::STRSTR_cpp(line, "::")) {
        ClassFactoryChunkI *parser_chunk = new ClassFactoryChunkI();
        ClassFactoryParser::parse(line, parser_chunk);

        for (int b=0;b<parser_chunk->getNumBlocks();b++) {
          ClassFactoryBlock *block = parser_chunk->enumBlock(b);
          if (block->getBlockType() == CF_BLOCKTYPE_METHOD) {
            ClassFactoryMethod *parser_method = block->getMethod();
            // find the method in the class xhd
            for (int cb=0;cb<chunk->getNumBlocks();cb++) {
              ClassFactoryBlock *cblock = chunk->enumBlock(cb);
              if (cblock->getBlockType() == CF_BLOCKTYPE_METHOD) {
                ClassFactoryMethod *class_method = cblock->getMethod();
                // test method name
                if (!STREQLSAFE(class_method->getName(), parser_method->getName())) continue;
                // test number of params
                if (parser_method->getNumParams() != class_method->getNumParams()) continue;
                // test method parameter types
                int doesntmatch = 0;
                for (int p=0;p<parser_method->getNumParams()&&!doesntmatch;p++)
                  if (!STREQLSAFE(parser_method->enumParam(p)->getType(), class_method->enumParam(p)->getType()))
                    doesntmatch=1;
                // seems to be the right method, set it as last parsed
                lastmethodparsed = class_method;
              }
            }
          }
        }
      }
    }

    ClassFactoryMethod *lm = currentmethod;
    if (gbracecount == 0 && currentmethod) currentmethod = NULL;
    if (lastmethodparsed && gbracecount > 0) currentmethod = lastmethodparsed;

    if (STRSTR(line, "//<?") || STRSTR(line, "/*<?")) {
      String buf = line;
      inlinefactory.processInlineXml(chunk, lm, buf, g_workspace);
      if (outf) Vfputs(buf.getNonConstVal(), outf);
    } else {
      if (outf) Vfputs(line, outf);
    }
  }

  Vfclose(inf);

  if (outf) {
    Vfclose(outf);

    unsigned long origcrc = crc32file(pathfile);
    unsigned long newcrc = crc32file(outfile);
    if (newcrc == origcrc) {
      unlink(outfile);
      return 2;
    }

    String oldfile = pathfile + String(".old");
    if (!ACCESS(oldfile, 0)) {
      #ifdef WIN32
      chmod(oldfile, _S_IREAD | _S_IWRITE);
      #else
      chmod(oldfile, 777);
      #endif
      unlink(oldfile);
    }
    rename(pathfile, oldfile);
    rename(path + filename + String("._tmp_"), pathfile);
  }
  return 1;
  PR_LEAVE();
}
