#include "precomp.h"
#include <bfc/string/stringdict.h>
#include <bfc/parse/paramparser.h>
#include "../classhierarchy/ClassFactoryChunk.h"
#include "../classhierarchy/ClassFactoryMethod.h"
#include "../classhierarchy/ClassFactoryParam.h"
#include "../classhierarchy/ClassFactoryBlock.h"
#include "../classhierarchy/ClassFactoryCodeBlock.h"
#include "../classhierarchy/XHDProcessor.h"
#include "../classhierarchy/Project.h"
#include "../classhierarchy/ClassFactoryParser.h"
#include "InlineXmlFactory.h"
#include "CPPToXHDMerge.h"
#include "../classhierarchy/Workspace.h"

#include "OutputFactoryBase.h"
#include <bfc/string/bigstring.h>
#include "Vfs.h"

#define IFBUF 8162

BEGIN_STRINGDICTIONARY(_cpp_typelist)
SDI("int", 0);
SDI("float", 0);
SDI("double", 0);
SDI("boolean", 0);
SDI("byte", 0);
SDI("hyper", 0);
SDI("long", 0);
SDI("short", 0);
SDI("small", 0);
SDI("char", 0);
END_STRINGDICTIONARY(_cpp_typelist, outputfactorybase_typelist);

BEGIN_STRINGDICTIONARY(_std_typelist)
SDI("time_t", 0);
SDI("FILE", 0);
SDI("HWND", 0);
SDI("HINSTANCE", 0);
SDI("HMODULE", 0);
SDI("HANDLE", 0);
SDI("HGDIOBJ", 0);
SDI("HRGN", 0);
SDI("HBITMAP", 0);
SDI("HCURSOR", 0);
SDI("HICON", 0);
SDI("HBRUSH", 0);
END_STRINGDICTIONARY(_std_typelist, outputfactorybase_stdtypelist);

OutputFactoryBase::OutputFactoryBase() {
  inlinexmlcount++;
  if (inlinexmlcount == 1) {
    inlinexml = new InlineXmlFactory();
  }

  indent_depth = default_indent_depth;
  use_col_braces = default_use_col_braces;
  verbose_comments = default_verbose_comments;
  eol_type = default_eol_type;
  lowercase_files = default_lowercase_files;
  indent_level = 0;
}

OutputFactoryBase::~OutputFactoryBase() {
  inlinexmlcount--;
  if (inlinexmlcount == 0) {
    delete inlinexml;
    inlinexml = NULL;
  }
}

void OutputFactoryBase::setChunk(ClassFactoryChunk *chunk) {
}

void OutputFactoryBase::_methodToText(ClassFactoryChunk *chunk, ClassFactoryMethod *m, String &s, int wantdefvals, const char *prefix, const char *torname, int wantimplementation) {
  BigString str;
  const char *rt = m->getReturnType();
  str += rt;
  int c = str.lastChar();
  if (c != '*' && c != '&' && *rt != 0) str += " ";
  if (prefix != NULL) str += prefix;


  String mname = m->getName();

  if (torname != NULL) {
    if (chunk->isConstructor(m))  {
      mname = torname;
    } else if (chunk->isDestructor(m)) {
      mname = StringPrintf("~%s", torname);
    }
  }

  str += mname; str += "(";
  for (int j=0;j<m->getNumParams();j++) {
    ClassFactoryParam *p = m->enumParam(j);
    if (j > 0) str += ", ";
    str += p->getType();
    c = str.lastChar();
    if (c != '*' && c != '&') str += " ";
    str += p->getName();
    if (wantdefvals) {
      const char *d = p->getDefaultValue();
      if (d != NULL && *d != 0) {
        str += " = ";
        str += d;
      }
    }
  }
  str += ")";
  s = str;
}

int OutputFactoryBase::needFriend() {
  return 0;
}

void OutputFactoryBase::setWorkspace(Workspace *workspace) {
  m_workspace = workspace;
}

Workspace *OutputFactoryBase::getWorkspace() {
  return m_workspace;
}

ClassFactoryChunk *OutputFactoryBase::findChunk(const char *classname) {
  if (m_workspace == NULL) return NULL;
  return m_workspace->findChunk(classname);
}

void OutputFactoryBase::processInlineXml(ClassFactoryChunk *chunk, ClassFactoryMethod *method, String &str, Workspace *workspace) {
  VFILE *fin = Vfopen_mem(str.getValue(), "rb", str.len());
  int wasoneline=0;
  String rebuild="";
  String xml;
  int inxml = 0;
  int inmlxml = 0;
  char *p, *d;
  char buf[IFBUF]="";
  while (!Vfeof(fin)) {
    Vfgets(buf, IFBUF, fin);
    buf[IFBUF-1]=0;
    p = CPPToXHDMerge::STRSTR_cpp(buf, "//<?");
    if (p != NULL) {
      inxml++;
      p += 4;
      rebuild.catn(buf, p-buf);
      while (*p && *p == ' ') p++;
      xml = p;
      rebuild.cat(xml);
      if (!xml.isempty()) {
        String s;
        doInline(chunk, method, xml, s, workspace);
        if (method != NULL)
          rebuild.cat("  ");
        rebuild.cat(s);
        rebuild.cat("\n");
      }
      continue;
    }
    p = CPPToXHDMerge::STRSTR_cpp(buf, "//?>");
    if (p != NULL) {
      inxml--;
      rebuild.cat(buf);
      continue;
    }
    if (inmlxml) {
      d = CPPToXHDMerge::STRSTR_cpp(buf, "*/");
      p = buf;
      goto commentterm;
    }
    p = CPPToXHDMerge::STRSTR_cpp(buf, "/*<?");
    if (p != NULL) {
      wasoneline = 1;
      inxml++;
      p += 4;
      d = STRSTR(p, "*/");
      commentterm:
      if (d == NULL) { // no terminator, grab end of line as xml
        xml.cat(p);
        d += 2;
        rebuild.cat(buf);
        inmlxml = 1; // continue grabing xml
        wasoneline = 1;
        continue;
      } else { // terminator found
        rebuild.catn(buf, (d+2)-buf);
        xml.catn(p, d-p); // copy remaining xml
        inmlxml = 0; // stop grabbing xml
        // xml is finished, process
        String s;
        doInline(chunk, method, xml, s, workspace);

        d += 2; // advance after term
        p = STRSTR(d, "/*?>*/"); // look for end of generated code

        if (p == NULL)
          rebuild.cat("\n");

        // inject code
        rebuild.cat(s);

        if (p != NULL) { // it's there, skip what's in between the tags, and grab the remaining
          rebuild.cat(p);
          p += 6;
          inxml--;
          xml.trunc(0);
          continue;
        } else {
          //rebuild.cat("\n");
          wasoneline = 0;
          continue;
        }
      }
    }
    if (inxml) {
      p = STRSTR(buf, "/*?>*/");
      if (p != NULL) {
        inxml--;
        if (wasoneline) rebuild.cat("\n");
        rebuild.cat(p);
        xml.trunc(0);
        continue;
      }
    }
    if (!inxml)
      rebuild.cat(buf);
  }
  Vfclose(fin);
  str = rebuild;
}

void OutputFactoryBase::doInline(ClassFactoryChunk *chunk, ClassFactoryMethod *method, String &source, String &dest, Workspace *workspace) {
  //XHDProcessor preproc(0, chunk);
  String sa = inlinexml->m_originalAncestor;
  String ca = chunk->getAncestor();
  inlinexml->m_originalAncestor = NULL;
  inlinexml->unloadAllTemplates();
  inlinexml->setChunk(chunk);
  inlinexml->setMethod(method);
  inlinexml->setWorkspace(workspace);
  inlinexml->insertFormatTemplate("inlinexml", source);
  inlinexml->processFormat();
  dest = inlinexml->enumModuleData(0);
  inlinexml->m_originalAncestor = sa;
  chunk->setAncestor(ca);
}

int OutputFactoryBase::isCallback(ClassFactoryMethod *method) {
  if (method != NULL) {
    const char *name = method->getName();
    if ( (STRLEN(name) > 2) && (name[0] == 'o') && (name[1] == 'n') && (isupper(name[2])) )
      return 1;
  }
  return 0;
}

void OutputFactoryBase::methodToText(ClassFactoryChunk *chunk, ClassFactoryMethod *m, String &s, int wantdefvals, const char *prefix, const char *torname, int wantimplementation) {
  _methodToText(chunk, m, s, wantdefvals, prefix, torname, wantimplementation);
}

void OutputFactoryBase::getDefaultForType(const char *type, String &s) {
  ParamParser pp(type, " ");
  const char *t = NULL;
  if (pp.getNumItems() == 1) { // simple type
    t = pp.enumItem(0);
  } else if (pp.getNumItems() == 2) { // either const type or type */&
    if (STRCHR(type, '*') || STRCHR(type, '&')) { // type */&

      t = pp.enumItem(0);
    } else { // const type
      t = pp.enumItem(1);
    }
  } else if (pp.getNumItems() == 3) { // const type */&
    t = pp.enumItem(1);
  } else {
    s = "NULL /*EDIT_ME!*/";
    return;
  }
  int a = outputfactorybase_typelist.getId(t);
  if (a != -1) {
    s = StringPrintf("(%s)0", type);
    return;
  } else {
    if (STRCASEEQL(t, "GUID")) {
      s = "INVALID_GUID";
    } else
      s = StringPrintf("(%s)NULL", type);
  }
}

String OutputFactoryBase::getBaseType(const String &verboseType) {
  String splice, retval = verboseType;
  // strip '*' and '&' and then any ending whitespace.
  do {
    splice = retval.rSpliceChar('*');
  } while (!retval.isempty());
  retval = splice;
  do {
    splice = retval.rSpliceChar('&');
  } while (!retval.isempty());
  retval = splice;
  while (retval.lastChar() == ' ') {
    retval.trunc(-1);
  }
  return retval;
}

int OutputFactoryBase::isClass(const char *stype, String &classname) {
  if (!stype || !*stype) return 0;
  ParamParser pp(stype, " ");
  const char *t = NULL;
  if (pp.getNumItems() == 1) { // simple type
    t = pp.enumItem(0);
  } else {
    for (int i=0;i<pp.getNumItems();i++) {
      const char *s = pp.enumItem(i);
      if (*s != '*' && *s != '&') {
        if (STRCASEEQL(s, "struct")) return 0;
        if (STRCASEEQL(s, "enum")) return 0;
        t = pp.enumItem(i);
      } else {
        break;
      }
    }
  }
  if (STRCHR(t, '<')) return 0;
  if (STRCHR(t, '>')) return 0;
  if (STRSTR(t, "struct")) return 0;
  if (STRSTR(t, "enum")) return 0;
  if (outputfactorybase_stdtypelist.getId(t) >= 0) return 0;
  if (isVoid(t)) return 0;
  classname = t;
  int a = outputfactorybase_typelist.getId(classname);
  if (a == -1) {
    // todo: check for exclusion
    return 1;
  }
  classname = "";
  return 0;
}

int OutputFactoryBase::tryFormat(const char *format) {
  return 0;
}

void OutputFactoryBase::setIndentDepth(int depth) {
  indent_depth = depth;
}

int OutputFactoryBase::getIndentDepth() {
  return indent_depth;
}

void OutputFactoryBase::setColumnarBraces(int braces) {
  use_col_braces = braces;
}

int OutputFactoryBase::getColumnarBraces() {
  return use_col_braces;
}

void OutputFactoryBase::setVerboseComments(int comments) {
  verbose_comments = comments;
}

int OutputFactoryBase::getVerboseComments() {
  return verbose_comments;
}

void OutputFactoryBase::setEOLType(int eol) {
  eol_type = eol;
}

int OutputFactoryBase::getEOLType() {
  return eol_type;
}

void OutputFactoryBase::setLowercaseFiles(int lowercase) {
  lowercase_files = lowercase;
}

int OutputFactoryBase::getLowercaseFiles() {
  return lowercase_files;
}

void OutputFactoryBase::resetIndentLevel() {
  indent_level = 0;
}

void OutputFactoryBase::setIndentLevel(int level) {
  indent_level = level;
}

void OutputFactoryBase::pushIndentLevel() {
  ++indent_level;
}

void OutputFactoryBase::popIndentLevel() {
  --indent_level;
}

int OutputFactoryBase::getIndentLevel() {
  return indent_level;
}

InlineXmlFactory * OutputFactoryBase::inlinexml = NULL;
int OutputFactoryBase::inlinexmlcount = 0;

