#ifndef _OUTPUTFACTORYBASE_H
#define _OUTPUTFACTORYBASE_H

#include "svc_srcClassFactoryI.h"

// -----------------------------------------------------------------------
// This header defines a bunch of things for which it would be really bad
// to compile into the general codebase.  It is meant only to be included
// in the scope of defining a service instance for svc_srcClassFactoryI
// -----------------------------------------------------------------------

/*<?<autoheader/>*/
class ClassFactoryChunk;
class ClassFactoryMethod;
class InlineXmlFactory;
class Workspace;
/*?>*/

class InlineXmlFactory;

// -----------------------------------------------------------------------
class EOLType {
  public:
    enum {
      EOL_WIN,
      EOL_MAC,
      EOL_LINUX,
    };
};

// -----------------------------------------------------------------------
class OutputFactoryBase : public svc_srcClassFactoryI {
public:
  OutputFactoryBase(); // just sets the defaults.
  virtual ~OutputFactoryBase();

  static const char *getServiceName() { return "Script Object Class Factory Service"; }

    // The actual service methods the instance needs to fill out.
    // We're just a helper object.  :)
  virtual int getNumAvailableFormats() { return 1; }
  virtual const char *enumAvailableFormat(int n) { if (n == 0) return getName(); return NULL; }
  virtual int tryFormat(const char *format);
  virtual const char *getName() { return NULL; }
  virtual const char *getDescription() { return NULL; }
  virtual int getNumModules() { return 0; }
  virtual const char *enumModuleName(int i) { return NULL; }
  virtual const char *enumModuleData(int i) { return NULL; }
  virtual void setChunk(ClassFactoryChunk *chunk);
  virtual int process() { return 0; }
  static void _methodToText(ClassFactoryChunk *chunk, ClassFactoryMethod *m, String &s, int wantdefvals, const char *prefix, const char *torname = NULL, int wantimplementation = 0);
  virtual int needFriend();

    // Simple support methods
  virtual const char *crlf();
  virtual String indent();
  virtual String openBrace();
  virtual const String &permitComment(const String &in);
  virtual void setWorkspace(Workspace *workspace);
  virtual Workspace *getWorkspace();
  virtual ClassFactoryChunk *findChunk(const char *classname);

  static void processInlineXml(ClassFactoryChunk *chunk, ClassFactoryMethod *method, String &str, Workspace *workspace=NULL);
  static void doInline(ClassFactoryChunk *chunk, ClassFactoryMethod *method, String &source, String &dest, Workspace *workspace=NULL);

protected:

    // Complex support methods
  int isCallback(ClassFactoryMethod *method); // return true if a callback method.
  void methodToText(ClassFactoryChunk *chunk, ClassFactoryMethod *m, String &s, int wantdefvals, const char *prefix, const char *torname = NULL, int wantimplementation = 0);
  void getDefaultForType(const char *type, String &s);
  String getBaseType(const String &verboseType);
  int isVoid(const char *stype) { return STRCASEEQLSAFE(stype, "void"); }
  int isClass(const char *stype, String &classname);

    // Parameter methods
  virtual void setIndentDepth(int depth);
  virtual int getIndentDepth();
  virtual void setColumnarBraces(int braces);
  virtual int getColumnarBraces();
  virtual void setVerboseComments(int comments);
  virtual int getVerboseComments();
  virtual void setEOLType(int eol);
  virtual int getEOLType();
  virtual void setLowercaseFiles(int lowercase);
  virtual int getLowercaseFiles();

    // State methods
  virtual void resetIndentLevel();
  virtual void setIndentLevel(int level);
  virtual void pushIndentLevel();
  virtual void popIndentLevel();
  virtual int getIndentLevel();

private:
  static InlineXmlFactory * inlinexml;
  static int inlinexmlcount;

    // Optional parameters.
  int indent_depth;
  int use_col_braces;
  int verbose_comments;
  int eol_type;
  int lowercase_files;

    // Default values for optiona
    enum {
      default_indent_depth     = 2,
      default_use_col_braces   = 0,
      default_verbose_comments = 1,
      default_eol_type         = EOLType::EOL_LINUX,
      default_lowercase_files  = 1,
    } default_values;

    // Runtime values
  int indent_level;
  Workspace *m_workspace;
};

//    If you're not aware of the limitations on this method, you
//    might be in for a world of hurt.  Just remember that the
//    returned char pointer will fall out of scope and be made
//    invalid on the line AFTER the function call which returns the
//    String.
//
//  CORRECT:
//
//      String val = indent();        // This copies the return into a new string
//
//      someCharPtrMethod(indent());  // This takes the const char * value and sends
//                                    // it into the given method.  That method better
//                                    // not assume the pointer will be valid outside
//                                    // of the method's scope.
//
//  INCORRECT:
//
//      const char *ptr = indent();   // This is dangerous....
//      someCharPtrMethod(ptr);       // Because the data is no longer valid here.
//
//  If you don't understand this, don't use the macros, just write your own codegen.
//

// ----------------------------------------------------------
// Right-hand string methods are useful.
// This allows the following:
//
//  String suffix = ".cpp";
//  String miscfile = "myfilename" + suffix;
//
static inline String operator + (const char *v1, const String &v2) {
  return String(v1) + v2;
}

// ----------------------------------------------------------
// Support for multiple CRLF formats.
static const char winEOL[] = { '\r', '\n', 0 }; // CR + LF + NUL
static const char macEOL[] = { 13, 0 };       // CR + NUL
static const char linEOL[] = { 10, 0 };       // LF + NUL
static const char *defaultEOL = linEOL;
// Return the string for the EOL, depending upon the target definition.
inline const char *OutputFactoryBase::crlf() {
  static const char *wEOL = winEOL;
  static const char *mEOL = macEOL;
  static const char *lEOL = linEOL;
  switch (eol_type) {
    case EOLType::EOL_WIN:
      return wEOL;
    case EOLType::EOL_MAC:
      return mEOL;
    case EOLType::EOL_LINUX:
      return lEOL;
  }
  return defaultEOL;
}

// ----------------------------------------------------------
// Return a string with the proper number of spaces, based on the
// current state of the indent_level and the desired indent_depth
inline String OutputFactoryBase::indent() {
  int i, j;
  String retval;
  for (i = 0; i < indent_level; i++) {
    if (indent_depth == -1) {
      retval += "\t";
    } else {
      for (j = 0; j < indent_depth; j++) {
        retval += " ";
      }
    }
  }
  return retval;
}

// ----------------------------------------------------------
// Return a string with the proper brace formatting, based on
// the state of use_col_braces
inline String OutputFactoryBase::openBrace() {
  String retval;
  if (use_col_braces) {
    retval = String(crlf()) + indent() + "{";
  } else {
    retval = " {";
  }
  return retval;
}

// ----------------------------------------------------------
// Passing a block of output data through this method will strip
// it if verbose comments are not enabled.
inline const String &OutputFactoryBase::permitComment(const String &in) {
  static const String blank = "";
  return (getVerboseComments()) ? in : blank;
}

// ----------------------------------------------------------
// "Readability" macros.  Basically, I'm turning all of these
// nice happy pretty functions into a set of code generation
// "opcodes" which represent archtypal strings that are used
// for most common types of codegen operations.

// "Carriage Return/Line Feed" (EOL)
#define CRLF crlf()

// "Open Brace"
#define OBRC openBrace()

// "Close Brace"
#define CBRC "}"

// "Indent"
#define INDT indent()

// "Push Indent Level"
#define PUSHI {pushIndentLevel();}

// "Pop Indent Level"
#define POPI {popIndentLevel();}

// "Verbose Comment Line"
#define CMMT(s) permitComment(s)

// "Start Verbose Comment Block"
#define START_COMMENT  if (getVerboseComments()) {

// "End Verbose Comment Block"
#define END_COMMENT  }

#endif
