  #include "precomp.h"

#ifdef WIN32
#include "resource.h"
#include <commctrl.h>
#endif

#include <api/apiinit.h>
#include <bfc/parse/paramparser.h>
#include <bfc/string/string.h>
#include <bfc/string/stringdict.h>
#include "outputfactories/svc_srcClassFactory.h"
#include "outputfactories/XHDWriter.h"
#include "classhierarchy/ClassScanner.h"
#include "inputfactories/XMLToChunk.h"
#include "outputfactories/CPPToXHDMerge.h"
#include "outputfactories/ImplementationFactory.h"
#include "outputfactories/InlineInterpreter.h"
#include "pworkspace.h"
#include "pproject.h"
#include <bfc/util/profiler.h>
#include <api/application/version.h>

#ifndef VERSION
#define VERSION "0.3.8"
#pragma WARNING("VERSION is " VERSION)
#endif

DECLARE_MODULE_SVCMGR;

// {F11F8398-0191-43da-84D0-4008BBDC084A}
static const GUID TL_PREPROC_GUID =
{ 0xf11f8398, 0x191, 0x43da, { 0x84, 0xd0, 0x40, 0x8, 0xbb, 0xdc, 0x8, 0x4a } };

enum CMDLINE_PARAM {
  CMDLINE_INPUTFILE,
  CMDLINE_HELP,
  CMDLINE_TEMPLATEDIR,
  CMDLINE_OUTPUTDIR,
  CMDLINE_YES,
  CMDLINE_NOLIST,
  CMDLINE_QUIET,
  CMDLINE_VERYQUIET,
  CMDLINE_NOPREPROC,
  CMDLINE_NOPARSE,
  CMDLINE_DEFINE,
  CMDLINE_NEWXHD,
  CMDLINE_STDIN,
  CMDLINE_FILE,
  CMDLINE_FILELIST,
  CMDLINE_VERSION,
};

BEGIN_STRINGDICTIONARY(_cmdlineswitches)
SDI("i",        CMDLINE_INPUTFILE);
SDI("stdin",    CMDLINE_STDIN);
SDI("outputdir",CMDLINE_OUTPUTDIR);
SDI("define",   CMDLINE_DEFINE);
SDI("d",        CMDLINE_DEFINE);
SDI("o",        CMDLINE_OUTPUTDIR);
SDI("t",        CMDLINE_TEMPLATEDIR);
SDI("yes",      CMDLINE_YES);
SDI("y",        CMDLINE_YES);
SDI("?",        CMDLINE_HELP);
SDI("h",        CMDLINE_HELP);
SDI("q",        CMDLINE_QUIET);
SDI("vq",       CMDLINE_VERYQUIET);
SDI("npp",      CMDLINE_NOPREPROC);
SDI("np",       CMDLINE_NOPARSE);
SDI("help",     CMDLINE_HELP);
SDI("file",     CMDLINE_FILE);
SDI("newxhd",   CMDLINE_NEWXHD)
SDI("fl",       CMDLINE_FILELIST);
SDI("filelist", CMDLINE_FILELIST)
SDI("version",  CMDLINE_VERSION)
END_STRINGDICTIONARY(_cmdlineswitches, cmdlineswitches)

String g_inputFile;
int    g_merge;
String g_xhdFile;
int    g_yes=0;
extern int g_noVar;
int    g_noList=0;
String g_projectFile;
PreprocProject *g_pp;
int cancel = 0;
String g_templateDir;
String g_outputDir;
int g_quiet = 0;
int g_veryquiet = 0;
int g_nopreproc = 0;
int g_noparse = 0;
int g_preprocstdin = 0;
String g_newclassdata;
String g_stdinclassname;
int m_projectdirty = 0;
String g_preprocfile;
String g_ppfileclassname;
String g_ppfileout;
String m_progressText;
String g_filelist;
PtrList<String> g_modulenames;

DWORD progress_start = 0;
void preprocessStdIn();
void preprocessFile();
void newXHDs(const char *data);
int xpp_confirmRewrite(const char *filename);

#ifndef WIN32
void clear_stdin(void);
#endif

#ifdef WIN32
HWND progressDlg = NULL;

BOOL CALLBACK ProgressDlgProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  switch(uMsg) {
    case WM_INITDIALOG: {
      EnableWindow(GetDlgItem(hwndDlg, IDOK), 0);
      SendMessage(GetDlgItem(progressDlg, IDC_PROGRESS), PBM_SETRANGE, 0, MAKELPARAM(0, 100));
      SendMessage(GetDlgItem(progressDlg, IDC_PROGRESS), PBM_SETPOS, 0, 0);
      return 1;
    }
    case WM_COMMAND:
      switch (LOWORD(wParam)) {
        case IDOK:
          EndDialog(hwndDlg, IDOK);
          return 0;
        case IDCANCEL:
          cancel = 1;
          EndDialog(hwndDlg, IDCANCEL);
          return 0;
      }
      break;
  }
  return NULL;
}

void xpp_updateProgress(int doupdate) {
  static DWORD last = 0;
  static int wi = 0;
  static char wheel[5] = "|/-\\";
  int sofar = Std::getTickCount() - progress_start;
  int s = sofar/1000;
  if (sofar > last + 100) doupdate = 1;
  if (doupdate) {
    last = sofar;
    SetDlgItemText(progressDlg, IDC_PROGRESS_TEXT, StringPrintf("%s... %c", m_progressText, wheel[wi]));
    SetDlgItemText(progressDlg, IDC_PROGRESS_TIME, StringPrintf("%2d:%02d", (s / 60), s % 60));
    UpdateWindow(progressDlg);
    wi = ++wi % 4;
  }
}
#else
void xpp_updateProgress(int doupdate) {
}
#endif

void writeModule(const char *filename, const char *data, const char *outputdir, ClassFactoryChunk *chunk) {
  String outpathfile = outputdir;
  outpathfile += filename;

  String hfile = chunk->getParams()->getItemValue("class_header");
  String cppfile = chunk->getParams()->getItemValue("class_module");
  if (hfile.iscaseequal(filename) || cppfile.iscaseequal(filename)) {
    if (!g_veryquiet) {
      printf(" - error: writing module would overwrite implementation\n");
      fflush(stdout);
    }
    return;
  }

  if (!xpp_confirmRewrite(outpathfile)) return;

  g_modulenames.addItem(new String(outpathfile));
  FILE *of = fopen(outpathfile, "wb");
  if (of == NULL) {
    if (!g_veryquiet) {
      printf("\nError, could not write %s", outpathfile.getValue());
      fflush(stdout);
    }
  } else {
    fwrite(data, STRLEN(data), 1, of);
    fclose(of);
  }
}

int didhelp = 0;

void setParam(const char *sw, const char *par) {
  int a = cmdlineswitches.getId(sw);
  switch (a) {
    case CMDLINE_INPUTFILE:
      g_inputFile = par;
      break;
    case CMDLINE_OUTPUTDIR:
      g_outputDir = par;
      if (!Std::isDirChar(g_outputDir.lastChar())) g_outputDir.cat(DIRCHARSTR);
      break;
    case CMDLINE_YES:
      g_yes = 1;
      break;
    case CMDLINE_QUIET:
      g_quiet = 1;
      break;
    case CMDLINE_STDIN:
      g_preprocstdin = 1;
      g_stdinclassname = par;
      break;
    case CMDLINE_TEMPLATEDIR:
      g_templateDir = par;
      break;
    case CMDLINE_FILE: {
        printf("-file param received with %s\n", par);
        fflush(stdout);
        ParamParser pp(par, ";,");
        if (pp.getNumItems() > 1) {
          g_preprocfile = pp.enumItem(0);
          g_ppfileout = pp.enumItem(1);
          if (pp.getNumItems() > 2) {
            g_ppfileclassname = pp.enumItem(2);
          }
        }
        printf("-file is %s %s %s\n", g_preprocfile.getValue(), g_ppfileout.getValue(), g_ppfileclassname.getValue());
        fflush(stdout);
      break;
      }
    case CMDLINE_VERYQUIET:
      g_quiet = 1;
      g_veryquiet = 1;
      break;
    case CMDLINE_NOPREPROC:
      g_nopreproc = 1;
      break;
    case CMDLINE_NOPARSE:
      g_noparse = 1;
      break;
    case CMDLINE_HELP: {
      didhelp = 1;
      String s = StringPrintf("xPP v%s\n\n"
                              "Usage:\n\nxpreproc -i:input.<xhd|xpp> will run a full preprocessing of the files referenced in input.<xhd|xpp>\n\n"
                              "You may use the following options and commands :\n\n"
                              "    -o or -outputdir:out_directory\n"
                              "                prepend output paths with out_directory/\n"
                              "    -t or -templatedir:template_directory\n"
                              "                add template_directory in search list\n"
                              "    -d or -define:symbol=value\n"
                              "                define global symbol to value\n"
                              "    -fl or -filelist:filename.txt\n"
                              "                writes the list of generated files in filename.txt\n"
                              "    -file:file_to_process[,classname]\n"
                              "                process a file with optional class context\n"
                              "    -stdin[:classname]\n"
                              "                process stdin with optional class context\n"
                              "    -q \n"
                              "                quiet, only output errors\n"
                              "    -vq\n"
                              "                very quiet, output only the result of -file/-stdin\n"
                              "    -np\n"
                              "                no parsing of source files (data may not be up to date)\n"
                              "    -npp\n"
                              "                no preprocessing (will not generate template targets)\n"
                              "    -y\n"
                              "                do not ask confirmation before overwriting target files\n"
                              "\n", VERSION);
      if (!g_quiet) {
        printf(s);
        fflush(stdout);
      }
      #ifdef WIN32
       MessageBox(NULL, s, "Help", 0);
      #endif
      break;
    }
    case CMDLINE_DEFINE: {
      ParamParser pp(par, ",;");
      for (int i=0;i<pp.getNumItems();i++) {
        ParamParser ppp(pp.enumItem(i), "=");
        if (ppp.getNumItems() == 2) {
          g_globalVars->setString(ppp.enumItem(0), ppp.enumItem(1));
        }
      }
      break;
    }
    case CMDLINE_NEWXHD:
      g_newclassdata = par;
      break;
    case CMDLINE_FILELIST:
      g_filelist = par;
      break;
    case CMDLINE_VERSION:
      printf(VERSION);
      exit(0);
      break;
  }
}

Workspace *g_workspace=NULL;
String *g_workspaceName;

ClassFactoryChunk *process(const char *file, int dowrite, const char *prfx);

#if defined(WIN32) & !defined(_CONSOLE)
int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
#else
int main(int argc, char **argv)
#endif
{
#if !defined(WIN32) | defined(_CONSOLE)
  String lpCmdLine;
  HINSTANCE hInstance = (HINSTANCE)getpid();
  for (int a=1;a<argc;a++) {
    if (!lpCmdLine.isempty()) lpCmdLine += " ";
    lpCmdLine += argv[a];
  }
#endif

  INIT_MODULE_SVCMGR();

  progress_start = Std::getTickCount();

  PR_ENTER("Full run");

  PreprocProject pw;
  g_pp = &pw;
  g_workspace = pw.getWorkspace();

  const char *outfile = NULL;

  WasabiVersion::setAppName("xpp");

  // Initialize The Library
  ApiInit::init(hInstance, TL_PREPROC_GUID, lpCmdLine);

  WASABI_API_SVC->service_register(new SrcClassFactoryCreatorSingle<ImplementationFactory>);

  // Parse command line
  ParamParser pp(lpCmdLine, " ");

  if (pp.getNumItems() == 0) {
    setParam("help", NULL);
  }

  for (int i=0;i<pp.getNumItems();i++) {
    String str = pp.enumItem(i);
    char *par = STRCHR(str, ':');
    char *cmd = str.getNonConstVal();
    if (*cmd == '-' && *(cmd+1) == '-') cmd++;
    cmd++;
    if (par) { *par = 0; par++; }
    setParam(cmd, par);
  }

  if (!didhelp) {
    if (!g_quiet) {
      printf("xPP v%s\n", VERSION);
      fflush(stdout);
    }
  }

  ((PreprocWorkspace *)g_workspace)->setTemplateDir(g_templateDir);

  if (g_inputFile.isempty()) {
    if (!didhelp) {
      if (!g_quiet) {
        printf("No input file specified, use /i:file.xhd or /i:file.xpp\n");
        fflush(stdout);
      }
    }
    return 0;
  }

  {
    int p = g_inputFile.rFindChar(".");
    String ext;
    if (p >= 0) ext = g_inputFile.getValue() + p + 1;
    if (ext.iscaseequal("xpp")) {
      if (!g_quiet) {
        printf("Processing project file %s\n", g_inputFile.getValue());
        fflush(stdout);
      }
      if (pw.loadFile(g_inputFile)) {
#ifdef WIN32
        progressDlg = CreateDialog(hInstance, MAKEINTRESOURCE(IDD_DIALOG_PROGRESS), NULL, ProgressDlgProc);
        SetWindowPos(progressDlg, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE|SWP_NOSIZE|SWP_SHOWWINDOW);
#endif
        int i, err=0;
        if (g_noparse && !g_quiet) {
          printf("No preparsing, using cached data, I cannot guarantee that this data is current\n");
          fflush(stdout);
        }
        for (i=0;i<pw.getNumFiles() && !err && !cancel;i++) {
#ifdef WIN32
          m_progressText = StringPrintf("Parsing %s", pw.enumFile(i));
          SendMessage(GetDlgItem(progressDlg, IDC_PROGRESS), PBM_SETPOS, (int)((double)(i+1)/pw.getNumFiles()*100.0), 0);
#endif
          xpp_updateProgress(1);
          if (process(pw.enumFile(i), 0, "  ") == NULL) { err++; }
          xpp_updateProgress(1);
        }
        if (!g_nopreproc) {
          for (i=0;i<pw.getNumFiles() && !err && !cancel;i++) {
            m_progressText = StringPrintf("Processing %s", pw.enumFile(i));
#ifdef WIN32
            SendMessage(GetDlgItem(progressDlg, IDC_PROGRESS), PBM_SETPOS, (int)((double)(i+1)/pw.getNumFiles()*100.0), 0);
#endif
            xpp_updateProgress(1);
            process(pw.enumFile(i), 1, "  ");
            xpp_updateProgress(1);
          }
        }
        if (cancel) {
          if (!g_veryquiet) {
            printf("aborted\n");
            fflush(stdout);
          }
        }
      }
    } else if (ext.iscaseequal("xhd")) {
      if (!g_quiet) {
        printf("Processing XHD file %s\n", g_inputFile.getValue());
        fflush(stdout);
      }
      int err=0;
      if (g_noparse && !g_quiet) {
        printf("No preparsing, using cached data, I cannot guarantee that this data is current\n");
        fflush(stdout);
      }
      if (process(g_inputFile, 0, "  ") == NULL) err++;
      if (cancel) {
        printf("aborted\n");
        fflush(stdout);
      } else if (!err && !g_nopreproc) {
        process(g_inputFile, 1, "  ");
      }
    }
  }

  if (g_newclassdata) {
    newXHDs(g_newclassdata);
  }

  if (g_preprocstdin) {
    preprocessStdIn();
  }

  if (!g_preprocfile.isempty()) {
    preprocessFile();
  }

  if (!g_filelist.isempty() && g_modulenames.getNumItems() > 0) {
    if (xpp_confirmRewrite(g_filelist)) {
      FILE *fl = fopen(g_filelist, "wt");
      if (!fl) {
        if (!g_veryquiet) {
          printf("\nerror opening %s for writing\n", g_filelist.getValue());
          fflush(stdout);
        }
      } else {
        foreach(g_modulenames)
          fprintf(fl, "%s\n", g_modulenames.getfor()->getValue());
        endfor
        FCLOSE(fl);
      }
    }
  }

  if (m_projectdirty) {
    if (!g_quiet) {
      printf("Saving xpp project file %s\n", g_inputFile.getValue());
      fflush(stdout);
    }
    pw.writeFile(g_inputFile);
  }

  DWORD total = Std::getTickCount() - progress_start;
  if (!g_quiet) {
    printf("\nTotal time : %dms\n", total);
    fflush(stdout);
  }

  return 0;
  PR_LEAVE();
}

ClassFactoryChunk *process(const char *file, int dowrite, const char *prfx) {
  PR_ENTER(StringPrintf("Processing file %s (%s)", file, dowrite ? "readwrite" : "readonly"));
  String inputFile = file;

  // create the XML to XHD parser class
  if (!Std::isDirChar(*inputFile.getValue()) && !STRCHR(inputFile, ':'))
    inputFile.prepend(g_pp->getProjectDir());

  if (dowrite) {
    if (!g_quiet) {
      printf("%sPreprocessing %s\n", prfx, inputFile.getValue());
      fflush(stdout);
    }
  }

  XMLToChunk r(inputFile);

  if (r.getNumChunks() == 0) {
    if (!g_quiet) {
      printf("%sNo XHD data to process\n", prfx);
      fflush(stdout);
    }
    return 0;
  }

  int numerrs = 0;

  PtrList<svc_srcClassFactory> factories;
  PtrList<String> formats;

  ClassFactoryChunkI *chunk = NULL;
  chunk = new ClassFactoryChunkI();
  chunk->copy(r.enumChunk(0));

  String xhd_dir_in="";
  String xhd_dir_out="";
  PathParser xp(file);
  if (Std::isDirChar(*file)) xhd_dir_in += DIRCHARSTR;
  for (int xi=0;xi<xp.getNumStrings()-1;xi++) {
    xhd_dir_in += xp.enumString(xi);
    xhd_dir_in += DIRCHARSTR;
  }
  xhd_dir_out = xhd_dir_in;
  if (!g_outputDir.isempty()) {
    if (!Std::isDirChar(*xhd_dir_out.getValue()) && !STRCHR(xhd_dir_out, ':'))
    xhd_dir_out.prepend(g_outputDir);
  }
  if (!Std::isDirChar(*xhd_dir_in.getValue()) && !STRCHR(xhd_dir_in, ':')) {
    xhd_dir_in.prepend(g_pp->getProjectDir());
    xhd_dir_out.prepend(g_pp->getProjectDir());
  }

  // process inline xml
  String hfile = chunk->getParams()->getItemValue("class_header");
  String cppfile = chunk->getParams()->getItemValue("class_module");
  if (!Std::isDirChar(*hfile.getValue()) && !STRCHR(hfile, ':')) {
    hfile.prepend(xhd_dir_in);
  }
  if (!Std::isDirChar(*cppfile.getValue()) && !STRCHR(cppfile, ':')) {
    cppfile.prepend(xhd_dir_in);
  }

  InlineInterpreter ii;
  if (!g_noparse) {
    ii.process(hfile, chunk, 0);
    ii.process(cppfile, chunk, 0);
  }

  // enumerate all factories in chunk, set up inheritance
  const char *xfactories = chunk->getParams()->getItemValue("xhd_factories");
  ParamParser pp(xfactories);
  int f, nf = pp.getNumItems();
  for (f=0;f<nf/*+1*/;f++) {
    String factory = /*(f == nf) ? "Implementation" : */pp.enumItem(f);
    //if (f < nf && factory.iscaseequal("Implementation")) continue; // Implementation must be last

    // find the factory
    SrcClassFactoryEnum e(factory);
    svc_srcClassFactory *cf = e.getFirst();
    if (cf != NULL) {
      cf->setChunk(chunk);
      if (!g_noparse) {
        cf->process();
      }
      factories.addItem(cf);
      formats.addItem(new String(factory));
    } else {
      if (dowrite) printf("template \"%s\" not found, ignoring\n", factory.getValue());
    }
  }

  if (!g_noparse) {
    PR_ENTER("CPP Merge");
    CPPToXHDMerge m(chunk, xhd_dir_in); // should be input dir
    PR_LEAVE();
    for (int ern=0;ern<chunk->getNumErrors();ern++) {
      int col;
      const char *ers = chunk->enumError(ern, &col);
      if (col >= 0) {
        if (!g_veryquiet) printf("%s%s(%d) : error : %s\n", prfx, inputFile.getValue(), col, ers);
      } else {
        if (!g_veryquiet) printf("%s%s(1) : error : %s\n", prfx, inputFile.getValue(), ers);
      }
      numerrs++;
      fflush(stdout);
    }
  }

  if (numerrs == 0 && dowrite) {

    int i, ni = factories.getNumItems();
    for (i=0;i<ni;i++) {
      svc_srcClassFactory *cf = factories.enumItem(i);
      if (formats.enumItem(i)->iscaseequal("Implementation")) continue;
      cf->setChunk(chunk);
      cf->process();
      for (int l=0;l<cf->getNumModules();l++) {
        const char *outfile = cf->enumModuleName(l);
        PR_ENTER(StringPrintf("Module %s", outfile));
        if (!g_noList && !g_quiet)
          printf("%s  %s -> %s%s (%s)", prfx, inputFile.getValue(), xhd_dir_out.getValue(), outfile, cf->getName());
        writeModule(outfile, cf->enumModuleData(l), xhd_dir_out, chunk);
        if (!g_noList && !g_quiet)
          printf("\n");
        fflush(stdout);
        PR_LEAVE();
      }
      WASABI_API_SVC->service_release(cf);
    }

    factories.removeAll();
    formats.deleteAll();

    if (!g_quiet) {
      printf("%sPreprocessing %s", prfx, hfile.getValue());
      fflush(stdout);
    }
    int _i = ii.process(hfile, chunk, 1);
    if (_i == 2) {
      if (!g_quiet) {
        printf(" (no change)"); printf("\n");
        fflush(stdout);
      }
    }
    if (!g_quiet) {
      printf("%sPreprocessing %s", prfx, cppfile.getValue());
      fflush(stdout);
    }
    _i = ii.process(cppfile, chunk, 1);
    if (_i == 2) {
      if (!g_quiet) {
        printf(" (no change)"); printf("\n");
        fflush(stdout);
      }
    }

    // rewrite XHD
    chunk->dropCodeBlocks(NULL, 1);
    chunk->getParams()->removeItem("friends");
    chunk->setAncestor(NULL);
    XHDWriter::convert(chunk, inputFile);
  } else if (numerrs > 0) {
    if (!g_veryquiet) {
      printf(StringPrintf("%s %d errors, code generation aborted.\n", prfx, numerrs));
      fflush(stdout);
    }
    return NULL;
  }
  ClassFactoryChunk *cc = g_workspace->findChunk(chunk->getClassName());
  if (!cc) { g_workspace->addChunk(chunk); }
  else { cc->copy(chunk); delete chunk; chunk = static_cast<ClassFactoryChunkI*>(cc); }
  return chunk;
  PR_LEAVE();
}

void xpp_errorMessage(const char *filename, int fileline, const char *incpath, int errcode, const char *errstr) {
  String file = incpath;
  if (!file.isempty() && !Std::isDirChar(file.lastChar())) file += Std::dirCharStr();
  file += filename;

  String s = StringPrintf("%s(%d) : xpp error", file.getValue(), fileline);
  if (errcode != -1) s += StringPrintf(" %d", errcode);
  s += StringPrintf(" : %s\n", errstr);
  //OutputDebugString(s);
  if (!g_veryquiet) {
    printf(s);
    fflush(stdout);
  }
  exit(errcode);
}

void xpp_warningMessage(const char *filename, int fileline, const char *incpath, int errcode, const char *errstr) {
  String file = incpath;
  if (!file.isempty() && !Std::isDirChar(file.lastChar())) file += Std::dirCharStr();
  file += filename;

  String s = StringPrintf("%s(%d) : xpp warning ", file.getValue(), fileline);
  if (errcode != -1) s += StringPrintf(" %d", errcode);
  s += StringPrintf(" : %s\n", errstr);
  //OutputDebugString(s);
  if (!g_quiet) {
    printf(s);
    fflush(stdout);
  }
}

#ifndef WIN32
void clear_stdin(void) {
    char thrash[80];
    fgets(thrash, 79, stdin);   /* gets will read characters from stdin until a newline (\n) is encountered and stored the string in thrash */
                    /* the newline (\n) is also read */
}
#endif

void preprocessHandle(FILE *fin, FILE *fout, String classname, char *ctxtname) {
  ClassFactoryChunk *context = NULL;
  if (classname.isempty()) {
    if (g_workspace->getNumChunks() == 0) {
      if (!g_veryquiet) {
        printf("no class available for %s context\n", ctxtname);
        fflush(stdout);
      }
    }
    context = g_workspace->enumChunk(0);
  } else {
    context = g_workspace->findChunk(classname);
    if (!context) {
      if (!g_veryquiet) {
        printf("%s context class %s not found\n", ctxtname, classname.getValue());
        fflush(stdout);
      }
    }
  }
  char buf[16384];
  BigString str;
  while (!feof(stdin)) {
    *buf = 0;
    fgets(buf, 16384, stdin); buf[16383] = 0;
    str += buf;
  }
  if (!str.isempty()) {
    InlineXmlFactory inlinefactory;
    String input = str.getValue();
    String result;
    inlinefactory.doInline(context, NULL, input, result, g_workspace);
    if (!result.isempty()) {
      fprintf(fout, result);
      fflush(fout);
    }
  }
}

void preprocessStdIn() {
  preprocessHandle(stdin, stdout, g_stdinclassname, "stdin");
}

void preprocessFile() {
  if (!g_quiet) {
    printf("Processing %s...\n", g_preprocfile.getValue());
    fflush(stdout);
  }
  if (access(g_preprocfile, 0)) {
    if (!g_veryquiet) {
      printf("\nfile %s specified in -file does not exist", g_preprocfile.getValue());
      fflush(stdout);
    }
    return;
  }
  if (!xpp_confirmRewrite(g_ppfileout)) return;
  FILE *fout = fopen(g_ppfileout, "wt");
  if (!fout) {
    printf("Could not open %s for writing, aborting -file processing", g_ppfileout.getValue());
    fflush(stdout);
    return;
  }
  FILE *fin = fopen(g_preprocfile, "rt");
  if (!fin) {
    printf("Could not open %s for reading, aborting -file processing", g_preprocfile.getValue());
    fflush(stdout);
    return;
  }
  preprocessHandle(fin, fout, g_stdinclassname, "-file");
  fclose(fout);

}

void newXHDs(const char *data) {
  m_projectdirty = 1;
}

int xpp_confirmRewrite(const char *filename) {
  if (!g_yes) {
    if (access(filename, 0) == 0) {
      #ifdef WIN32
        int a = MessageBox(NULL, StringPrintf("\nfile %s already exists, overwrite ?", filename), "Warning", MB_YESNO|MB_ICONWARNING);
        if (a == IDNO) return 0;
      #else
        printf(StringPrintf("\nfile %s already exists, overwrite ? (y/n/a[ll]) ", filename).getValue());
        fflush(stdout);
        char c;
        while (1) {
          scanf("%c", &c);
          clear_stdin();
          c = TOUPPER(c);
          if (c == 'Y') break;
          else if (c == 'N') return 0;
          else if (c == 'A') { g_yes = 1; break; }
        }
      #endif
    }
  }
  return 1;
}
